/*
** Module   : XNodeRefImpl
** Abstract : DOM XML Node reference resource type implementation.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------Description------------------------
** 001 EVL 20130204 Created initial version. Major rework to use w3c backend.
** 002 CA  20130221 Added LOCAL-NAME support. 
** 003 EVL 20130222 Adding NODE-VALUE attribute and SET-ATTRIBUTE-NODE method.
** 004 EVK 20130531 Adding NODE-VALUE attribute and GET-PARENT, GET-ATTRIBUTE-NODE, NORMALIZE
**                  LONGCHAR-TO-NODE-VALUE, NODE-VALUE-TO-LONGCHAR, CLONE-NODE methods.
*/

package com.goldencode.p2j.xml;

import org.w3c.dom.*;
import com.goldencode.p2j.util.*;

/**
 * Implementation of the X-Node-Reference specific features of the Progress
 * DOM XML support objects.
 */
public class XNodeRefImpl
extends XEntityImpl
implements XNodeRef
{

   public static final String RESOURCE_TYPE = "X-NODEREF";

   /**
    * Default constructor.
    */
   public XNodeRefImpl()
   {
      super();
      setResourceType();
   }

   /**
    * Special constructor accepting the namspace URI specification to make
    * another node reference case.
    *
    * @param    h
    *           The valid handle class instance of the document to be created.
    */
   public XNodeRefImpl(handle h)
   {
      super(h);
      setResourceType();
   }

   /**
    * Special constructor accepting the namspace URI specification to make
    * another node reference case.
    *
    * @param    node
    *           The node associated with this node.
    * @param    h
    *           The valid handle class instance of the document to be created.
    */
   public XNodeRefImpl(handle h, Node node)
   {
      super(h, node);
      setResourceType();
   }

   private void setResourceType()
   {
      setResourceType(RESOURCE_TYPE);
   }

   // Attributes
   /**
    * Getting the attribute names available for current node reference.
    * 
    * @return   The comma separated character string of all available
    *           attributed.
    */
   @Override
   public character getAttributeNames()
   {
      if (!valid())
      {
         return new character();
      }
      if (unknown())
      {
         invalidNodeError("ATTRIBUTE-NAMES");
         return new character();
      }

      final StringBuilder sb = new StringBuilder();
      final NamedNodeMap nnm = getNode().getAttributes();
      if (nnm == null)
      {
         return new character("");
      }
      final int iSize = nnm.getLength();
      for (int i = 0; i < iSize; i++)
      {
         sb.append(((Attr)nnm.item(i)).getName());
         if (i < iSize - 1)
         {
            sb.append(',');
         }
      }   

      return new character(sb.toString());
   }

   /**
    * Getting the value of the XML node value attribute.
    * 
    * @return   The character value of the node reference.
    */
   @Override
   public character getNodeValue()
   {
      if (!valid())
      {
         return new character();
      }
      if (unknown())
      {
         invalidNodeError("NODE-VALUE");
         return new character();
      }
      try
      {
         final String nodeValue = getNode().getNodeValue();
         return new character(nodeValue == null ? "" : nodeValue);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.DOMSTRING_SIZE_ERR)
         {
            commonDomError("NODE-VALUE", "Too long node-value");
         }
         return new character();
      }
   }
   
   /**
    * Setting the value of the XML node value attribute.
    * 
    * @param    value
    *           The new value of the node value attribute.
    */
   @Override
   public void setNodeValue(character value)
   {
      if (!valid())
      {
         return;
      }
      if (value.isUnknown())
      {
         invalidArgumentAssignError("NODE-VALUE");
         return;
      }
      if (unknown() || !valid())
      {
         final String[] errMsg =
            {
               "X-NODEREF must be associated with a valid X-DOCUMENT in order to use it in " +
                  "method NODE-VALUE",
               "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
            };
         final int[] errorCodes = { 9102, 3131 };
         ErrorManager.recordOrThrowError(errorCodes, errMsg, false);
         return;
      }

      final Node node = getNode();
      final int nodeType = node.getNodeType();
      if (nodeType == Node.ELEMENT_NODE || nodeType == Node.ENTITY_REFERENCE_NODE)
      {
         final String[] errMsg =
            {
               "X-NODEREF or X-DOCUMENT NODE-VALUE got an error: An " +
                  "attempt was made to modify an object where modifications are not allowed",
               "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
            };
         final int[] errorCodes = { 9082, 3131 };
         ErrorManager.recordOrThrowError(errorCodes, errMsg, false);
         return;
      }

      try
      {
         node.setNodeValue(value.toStringMessage());
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            final String[] errMsg =
               {
                  "X-NODEREF or X-DOCUMENT NODE-VALUE got an error: An " +
                     "attempt was made to modify an object where modifications are not allowed",
                  "Unable to set attribute NODE-VALUE in widget  of type X-NODEREF"
               };
            final int[] errorCodes = { 9082, 3131 };
            ErrorManager.recordOrThrowError(errorCodes, errMsg, false);
            return;
         }
      }
   }
   
   /**
    * Setting the value of the XML node value attribute.
    * 
    * @param    value
    *           The new value of the node value attribute.
    */
   @Override
   public void setNodeValue(String value)
   {
      setNodeValue(new character(value));
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param xNodeRef The valid handle class instance to use for the cloned XML node.
    * @param deep     A logical that if <code>true</code> specifies that the whole sub-trees is to be
    *                 copied. The default value is {@code false}.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical cloneNode(handle xNodeRef, logical deep)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("CLONE-NODE");
         return new logical(false);
      }

      if (!xNodeRef._isValid())
      {
         invalidArgumentError("CLONE-NODE");
         return new logical(false);
      }

      final Node node = getNode();
      final Node clonedNode = node.cloneNode(deep.booleanValue());
      xNodeRef.assign(new XNodeRefImpl(xNodeRef, clonedNode));

      return xNodeRef.isValid();
   }

   /**
    * Creating the new node within the current Document object.
    *
    * @param xNodeRef The valid handle class instance to use for the cloned XML node.
    * @param deep     A logical that if <code>true</code> specifies that the whole sub-trees is to
    *                 be copied. The default value is {@code false}.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical cloneNode(handle xNodeRef, boolean deep)
   {
      return cloneNode(xNodeRef, new logical(deep));
   }

   /**
    * Getting the Parent of current node.
    *
    * @param    hNoderefParent
    *           A valid handle to assign parent XML node.
    *
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical getParent(handle hNoderefParent)
   {
      if (!valid())
      {
         return new logical(false);
      }
      if (unknown())
      {
         invalidNodeError("GET-PARENT");
         return new logical(false);
      }
      if (hNoderefParent == null || !hNoderefParent._isValid())
      {
         invalidArgumentError("GET-PARENT");
         return new logical(false);
      }

      final Node node = getNode().getParentNode();

      hNoderefParent.assign(new XNodeRefImpl(hNoderefParent, node));

      return hNoderefParent.isValid();
   }

   // Methods
   /**
    * Deleting the internals of the current X-Node-Reference.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical deleteNode()
   {
      if (!valid())
      {
         return new logical(false);
      }
      if (unknown())
      {
         invalidNodeError("DELETE-NODE");
         return new logical(false);
      }
      final Node node = getNode();
      final Node parentNode = node.getParentNode();
      if (parentNode == null)
      {
         invalidNodeError("DELETE-NODE");
         return new logical(false);
      }
      try
      {
         parentNode.removeChild(node);
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NOT_FOUND_ERR)
         {
            invalidNodeError("DELETE-NODE");
         }
         else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            invalidNodeError("DELETE-NODE");
         }
         return new logical(false);
      }

      return new logical(true);
   }
   
   /**
    * Getting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to return.
    *
    * @return   The comma separated character string of all available
    *           attributed.
    */
   @Override
   public character getAttribute(String name)
   {
      return getAttribute(new character(name));
   }
   
   /**
    * Getting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to return.
    *
    * @return   Value as a character string, or the empty string
    *           if that attribute does not have a specified or default value.
    */
   @Override
   public character getAttribute(character name)
   {
      if (!valid())
      {
         return new character();
      }
      if (unknown())
      {
         invalidNodeError("GET-ATTRIBUTE");
         return new character();
      }
      if (name.isUnknown())
      {
         invalidArgumentError("GET-ATTRIBUTE");
         return new character();
      }
      final Node node = getNode();
      String result = null;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         final Element elementNode = (Element) node;
         result = elementNode.getAttribute(name.toStringMessage());
      }

      return new character((result == null) ? "" : result);
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(String name, String value)
   {
      return setAttribute(new character(name), new character(value));
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(String name, character value)
   {
      return setAttribute(new character(name), value);
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success
    *           <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(character name, String value)
   {
      return setAttribute(name, new character(value));
   }
   
   /**
    * Setting the attribute value for a given attribute name.
    * 
    * @param    name
    *           The name of the attribute to change.
    * @param    value
    *           The new value for the given attribute name.
    * 
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setAttribute(character name, character value)
   {
      if (!valid())
      {
         return new logical(false);
      }
      if (unknown())
      {
         invalidNodeError("SET-ATTRIBUTE");
         return new logical(false);
      }
      if (name.isUnknown() || value.isUnknown())
      {
         invalidArgumentError("SET-ATTRIBUTE");
         return new logical(false);
      }

      final Node node = getNode();
      final logical result;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         final Element elementNode = (Element) node;
         try
         {
            elementNode.setAttribute(name.toStringMessage(), value.toStringMessage());
         }
         catch (DOMException e)
         {
            if (e.code == DOMException.INVALID_CHARACTER_ERR)
            {
               commonDomError("SET-ATTRIBUTE", "An invalid character was specified");
            }
            else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
            {
               commonDomError("SET-ATTRIBUTE", "Read only node"); //TODO need correct message
            }
            return new logical(false);
         }
         result = new logical(true);//TODO simplify it
      }
      else
      {
         commonDomError("SET-ATTRIBUTE", "Invalid node type");
         result = new logical(false);
      }

      return result;
   }

   /**
    * Associates XML attribute node with the given X-Noderef object handle.
    * 
    * @param    handle
    *           The handle that represents an XML attribute node created with create node or
    *           create node namespace.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical setAttributeNode(handle handle)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (handle._isValid() && !handle.isUnknown() && handle.get() instanceof HandleResource)
      {
         final HandleResource resource = (HandleResource) handle.get();
         if (XDocumentImpl.TYPE_X_DOCUMENT.equals(resource.getResourceType().toStringMessage()))
         {
            invalidArgumentError("SET-ATTRIBUTE-NODE");
            return new logical(false);
         }
      }

      if (unknown() || handle.isUnknown())
      {
         invalidNodeError("SET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      final Object resource = handle.get();
      if (!(resource instanceof XNodeRefImpl))
      {
         invalidArgumentError("SET-ATTRIBUTE-NODE");
         return new logical(false);
      }
      final Node currentNode = getNode();
      final Node handleNode = ((XNodeRefImpl) resource).getNode();
      final logical result;
      if (currentNode.getNodeType() == Node.ELEMENT_NODE &&
              handleNode.getNodeType() == Node.ATTRIBUTE_NODE)
      {
         final Element elementNode = (Element) currentNode;
         try
         {
            elementNode.setAttributeNode((Attr) handleNode);
         }
         catch (DOMException e)
         {
            if (e.code == DOMException.WRONG_DOCUMENT_ERR)
            {
               invalidNodeError("SET-ATTRIBUTE-NODE");
            }
            else if (e.code == DOMException.INUSE_ATTRIBUTE_ERR)
            {
               commonDomError("SET-ATTRIBUTE-NODE", "Attribute already exists");
            }
            else if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
            {
               commonDomError("SET-ATTRIBUTE-NODE", "An attempt was made to modify an " +
                       "object where modifications are not allowed");
            }
            return new logical(false);
         }

         result = new logical(true);
      }
      else
      {
         commonDomError("SET-ATTRIBUTE-NODE", "Invalid node type");
         result = new logical(false);
      }

      return result;
   }

   /**
    * Associates XML attribute node with the specified name to the given X-Noderef object
    * handle.
    *
    * @param    handle
    *           The handle that represents an XML attribute node with the specified name.
    * @param    attributeName
    *           The string that represents an XML attribute node with the specified name.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   public logical getAttributeNode(handle handle, String attributeName)
   {
      return getAttributeNode(handle, new character(attributeName));
   }

   /**
    * Associates XML attribute node with the specified name to the given X-Noderef object
    * handle.
    *
    * @param    handle
    *           The handle that represents an XML attribute node with the specified name.
    * @param    attributeName
    *           The character that represents an XML attribute node with the specified name.
    *
    * @return   The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical getAttributeNode(handle handle, character attributeName)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (handle._isValid() && !handle.isUnknown() && handle.get() instanceof HandleResource)
      {
         final HandleResource resource = (HandleResource) handle.get();
         if (XDocumentImpl.TYPE_X_DOCUMENT.equals(resource.getResourceType()))
         {
            invalidArgumentError("GET-ATTRIBUTE-NODE");
            return new logical(false);
         }
      }

      if (!handle._isValid() || attributeName.isUnknown())
      {
         invalidArgumentError("GET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("GET-ATTRIBUTE-NODE");
         return new logical(false);
      }

      final Node node = getNode();
      final logical result;
      if (node.getNodeType() == Node.ELEMENT_NODE)
      {
         final Element elementNode = (Element) node;
         final Attr attrNode = elementNode.getAttributeNode(attributeName.toStringMessage());
         if (attrNode != null)
         {
            handle.assign(new XNodeRefImpl(handle, attrNode));
         }
         result = handle.isValid();
      }
      else
      {
         commonDomError("GET-ATTRIBUTE-NODE", "Invalid node type");
         result = new logical(false);
      }
      return result;
   }

   /**
    * Setting the binary value of the XML node value attribute.
    *
    * @param value The new value of the node value attribute.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical longcharToNodeValue(longchar value)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("LONGCHAR-TO-NODE-VALUE");
         return new logical(false);
      }

      if (value == null || value.isUnknown())
      {
         invalidArgumentError("LONGCHAR-TO-NODE-VALUE");
      }

      final Node node = getNode();

      node.setTextContent(value.toStringMessage());
      return new logical(true);
   }

   /**
    * Setting the XML node value attribute to binary value of longchar.
    *
    * @param value To save the value on XML node to this parameter.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical nodeValueToLongchar(longchar value)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("NODE-VALUE-TO-LONGCHAR");
         return new logical(false);
      }

      value.setValue(getNode().getNodeValue());

      return new logical(true);
   }

   /**
    * Setting the binary value of the XML node value attribute.
    *
    * @param value The new value of the node value attribute.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical memptrToNodeValue(memptr value)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("MEMPTR-TO-NODE-VALUE");
         return new logical(false);
      }

      if (value == null || value.isUnknown())
      {
         invalidArgumentError("MEMPTR-TO-NODE-VALUE");
      }

      final Node node = getNode();


      try
      {
         node.setTextContent(new String(value.getByteArray()));
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("MEMPTR-TO-NODE-VALUE", "Read only node");//TODO need correct message
            return new logical(false);
         }
      }
      return new logical(true);
   }

   /**
    * Setting the XML node value attribute to binary value of memptr.
    *
    * @param value To save the value on XML node to this parameter.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical nodeValueToMemptr(memptr value)
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("NODE-VALUE-TO-MEMPTR");
         return new logical(false);
      }

      value.setString(getNode().getNodeValue(), 1);

      return new logical(true);
   }

   /**
    * Normalizes TEXT and ATTRIBUTE nodes in the full depth of the sub-tree under this XML node.
    *
    * @return The <code>true</code> in case of success <code>false</code> otherwise.
    */
   @Override
   public logical normalize()
   {
      if (!valid())
      {
         return new logical(false);
      }

      if (unknown())
      {
         invalidNodeError("NORMALIZE");
         return new logical(false);
      }

      final Node node = getNode();
      short nodeType = node.getNodeType();
      if (nodeType == Node.ELEMENT_NODE)
      {
         node.normalize();
      }
      else
      {
         commonDomError("NORMALIZE", "Invalid node type");
         return new logical(false);
      }

      return new logical(true);
   }

   /**
    * Get the value of the LOCAL-NAME attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getLocalName()
   {
      if (!valid())
      {
         return new character();
      }

      if (unknown())
      {
         invalidNodeError("LOCAL-NAME");
         return new character();
      }


      final String localName = getNode().getLocalName();

      if (localName != null)
      {
         return new character(localName);
      }
      else
      {
         return new character("");
      }
   }

   /**
    * Setter for NAMESPACE-PREFIX attribute.
    *
    * @param    prefix
    *           The new value of the NAMESPACE-PREFIX attribute.
    */
   @Override
   public void setNamespacePrefix(character prefix)
   {
      if (!valid())
      {
         return;
      }
      if (unknown())
      {
         final String[] errMsg =
         {
            "X-NODEREF must be associated with a valid X-DOCUMENT in order " +
               "to use it in method NAMESPACE-PREFIX",
            "Unable to set attribute NAMESPACE-PREFIX in widget  of type X-NODEREF"
         };
         final int[] errorCodes = { 9102, 3131 };
         ErrorManager.recordOrThrowError(errorCodes, errMsg, false);
         return;
      }

      final Node currentNode = getNode();
      final String prefixStr = prefix.getValue(); //null is ok here
      try
      {
         if (currentNode.getNodeType() == Node.ELEMENT_NODE)
         {
            final Element elementNode = (Element) currentNode;
            elementNode.setPrefix(prefixStr);
         }
         else if (currentNode.getNodeType() == Node.ATTRIBUTE_NODE)
         {
            final Attr attributeNode = (Attr) currentNode;
            attributeNode.setPrefix(prefixStr);
         }
      }
      catch (DOMException e)
      {
         if (e.code == DOMException.NO_MODIFICATION_ALLOWED_ERR)
         {
            commonDomError("NAMESPACE-PREFIX", "An attempt was made to modify an " +
                    "object where modifications are not allowed");
         }
         else if (e.code == DOMException.NAMESPACE_ERR
                    || e.code == DOMException.INVALID_CHARACTER_ERR)
         {
            final String[] errMsg =
            {
               "X-NODEREF or X-DOCUMENT NAMESPACE-PREFIX got an error: " +
                  "XML Parser not initialized",
               "Unable to set attribute NAMESPACE-PREFIX in widget  of type X-NODEREF"
            };
            final int[] errorCodes = { 9082, 3131 };
            ErrorManager.recordOrThrowError(errorCodes, errMsg, false);
         }
      }
   }
}
